import json
import os
import requests
import numpy as np

from typing import List

from ophyd.areadetector import ADComponent as ADCpt, PilatusDetectorCam, DetectorBase
from ophyd.areadetector.plugins import FileBase

from bec_lib.core import BECMessage, MessageEndpoints
from bec_lib.core.file_utils import FileWriterMixin
from bec_lib.core import bec_logger

logger = bec_logger.logger


class PilatusDetectorCamEx(PilatusDetectorCam, FileBase):
    pass


class PilatusCsaxs(DetectorBase):
    """

    in device config, device_access needs to be set true to inject the device manager
    """

    _html_docs = ["PilatusDoc.html"]
    cam = ADCpt(PilatusDetectorCamEx, "cam1:")

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        device_manager=None,
        **kwargs,
    ):
        self.device_manager = device_manager
        self.username = "e21206"  # TODO get from config
        # self.username = self.device_manager.producer.get(MessageEndpoints.account()).decode()
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        # TODO how to get base_path
        self.service_cfg = {"base_path": f"/sls/X12SA/data/{self.username}/Data10/pilatus_2/"}
        self.filewriter = FileWriterMixin(self.service_cfg)
        self.num_frames = 0
        self.readout = 0.003  # 3 ms
        self.triggermode = 0  # 0 : internal, scan must set this if hardware triggered

    def _get_current_scan_msg(self) -> BECMessage.ScanStatusMessage:
        msg = self.device_manager.producer.get(MessageEndpoints.scan_status())
        return BECMessage.ScanStatusMessage.loads(msg)

    def stage(self) -> List[object]:
        # TODO remove
        # scan_msg = self._get_current_scan_msg()
        # self.metadata = {
        #     "scanID": scan_msg.content["scanID"],
        #     "RID": scan_msg.content["info"]["RID"],
        #     "queueID": scan_msg.content["info"]["queueID"],
        # }
        self.scan_number = 10  # scan_msg.content["info"]["scan_number"]
        self.exp_time = 0.5  # scan_msg.content["info"]["exp_time"]
        self.num_frames = 3  # scan_msg.content["info"]["num_points"]
        # TODO remove
        # self.username = self.device_manager.producer.get(MessageEndpoints.account()).decode()

        # set pilatus threshol
        self._set_threshold()

        # set Epic PVs for filewriting
        self.cam.file_path.set(f"/dev/shm/zmq/")
        self.cam.file_name.set(f"{self.username}_2_{self.scan_number:05d}")
        self.cam.auto_increment.set(1)  # auto increment
        self.cam.file_number.set(0)  # first iter
        self.cam.file_format.set(0)  # 0: TIFF
        self.cam.file_template.set("%s%s_%5.5d.cbf")

        # compile zmq stream for data transfer
        scan_dir = self.filewriter._get_scan_directory(
            scan_bundle=1000, scan_number=self.scan_number, leading_zeros=5
        )
        self.destination_path = os.path.join(
            self.service_cfg["base_path"]
        )  # os.path.join(self.service_cfg["base_path"], scan_dir)
        data_msg = {
            "source": [
                {
                    "searchPath": "/",
                    "searchPattern": "glob:*.cbf",
                    "destinationPath": self.destination_path,
                }
            ]
        }
        logger.info(data_msg)
        headers = {"Content-Type": "application/json", "Accept": "application/json"}

        res = requests.put(
            url="http://x12sa-pd-2:8080/stream/pilatus_2",
            data=json.dumps(data_msg),
            headers=headers,
        )

        if not res.ok:
            res.raise_for_status()

        # prepare writer
        data_msg = [
            "zmqWriter",
            self.username,
            {
                "addr": "tcp://x12sa-pd-2:8888",
                "dst": ["file"],
                "numFrm": self.num_frames,
                "timeout": 2000,
                "ifType": "PULL",
                "user": self.username,
            },
        ]

        res = requests.put(
            url="http://xbl-daq-34:8091/pilatus_2/run",
            data=json.dumps(data_msg),
            headers=headers,
        )

        if not res.ok:
            res.raise_for_status()

        self._set_acquisition_params(
            exp_time=self.exp_time,
            readout=self.readout,
            num_frames=self.num_frames,
            triggermode=self.triggermode,
        )

        return super().stage()

    def unstage(self) -> List[object]:
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        data_msg = [
            "zmqWriter",
            self.username,
            {
                "frmCnt": self.num_frames,
                "timeout": 2000,
                "ifType": "PULL",
                "user": self.username,
            },
        ]
        logger.info(data_msg)

        res = requests.put(
            url="http://xbl-daq-34:8091/pilatus_1/run",
            data=json.dumps(data_msg),
            headers=headers,
        )
        # Reset triggermode to internal
        self.triggermode = 0

        if not res.ok:
            res.raise_for_status()
        return super().unstage()

    def _set_threshold(self) -> None:
        # TODO readout mono, monitor threshold and set it if mokev is different
        # mokev = self.device_manager.devices.mokev.obj.read()["mokev"]["value"]
        # TODO remove
        mokev = 16
        # TODO refactor naming from name, pilatus_2
        pil_threshold = self.cam.threshold_energy.read()["pilatus_2_cam_threshold_energy"]["value"]
        if not np.isclose(mokev / 2, pil_threshold, rtol=0.05):
            self.cam.threshold_energy.set(mokev / 2)

    def _set_acquisition_params(
        self, exp_time: float, readout: float, num_frames: int, triggermode: int
    ) -> None:
        """set acquisition parameters on the detector

        Args:
            exp_time (float): exposure time
            readout (float): readout time
            num_frames (int): images per scan
            triggermode (int):
                0 Internal
                1 Ext. Enable
                2 Ext. Trigger
                3 Mult. Trigger
                4 Alignment
        Returns:
            None
        """
        self.cam.acquire_time.set(exp_time)
        self.cam.acquire_period.set(exp_time + readout)
        self.cam.num_images.set(num_frames)
        self.cam.num_exposures.set(1)
        self.cam.trigger_mode.set(triggermode)

    def acquire(self) -> None:
        self.cam.acquire.set(1)

    def stop(self, *, success=False) -> None:
        self.cam.acquire.set(0)
        super().stop(success=success)
        self._stopped = True



from setuptools import setup

__version__ = "0.4.0"

if __name__ == "__main__":
    setup(
        install_requires=["ophyd", "typeguard", "prettytable", "bec_lib", "numpy", "pyyaml"],
        extras_require={"dev": ["pytest", "pytest-random-order", "black"]},
        version=__version__,
    )



# -*- coding: utf-8 -*-
"""
Created on Tue Nov  9 16:12:47 2021

@author: mohacsi_i
"""

from ophyd import Device, Component, EpicsSignal, EpicsSignalRO, Kind
from ophyd import PVPositioner, Signal
from ophyd.pseudopos import (
    pseudo_position_argument,
    real_position_argument,
    PseudoSingle,
    PseudoPositioner,
)


class DelayStatic(Device):
    """Static axis for the T0 output channel
    It allows setting the logic levels, but the timing is fixed.
    The signal is high after receiving the trigger until the end
    of the holdoff period.
    """

    # Other channel stuff
    ttl_mode = Component(EpicsSignal, "OutputModeTtlSS.PROC", kind=Kind.config)
    nim_mode = Component(EpicsSignal, "OutputModeNimSS.PROC", kind=Kind.config)
    polarity = Component(
        EpicsSignal,
        "OutputPolarityBI",
        write_pv="OutputPolarityBO",
        name="polarity",
        kind=Kind.config,
    )
    amplitude = Component(
        EpicsSignal, "OutputAmpAI", write_pv="OutputAmpAO", name="amplitude", kind=Kind.config
    )
    offset = Component(
        EpicsSignal, "OutputOffsetAI", write_pv="OutputOffsetAO", name="offset", kind=Kind.config
    )


class DummyPositioner(PVPositioner):
    setpoint = Component(EpicsSignal, "DelayAO", put_complete=True, kind=Kind.config)
    readback = Component(EpicsSignalRO, "DelayAI", kind=Kind.config)
    done = Component(Signal, value=1)
    reference = Component(EpicsSignal, "ReferenceMO", put_complete=True, kind=Kind.config)


class DelayPair(PseudoPositioner):
    """Delay pair interface for DG645

    Virtual motor interface to a pair of signals (on the frontpanel).
    It offers a simple delay and pulse width interface for scanning.
    """

    # The pseudo positioner axes
    delay = Component(PseudoSingle, limits=(0, 2000.0), name="delay")
    width = Component(PseudoSingle, limits=(0, 2000.0), name="pulsewidth")
    # The real delay axes
    # ch1 = Component(EpicsSignal, "DelayAI", write_pv="DelayAO", name="ch1", put_complete=True, kind=Kind.config)
    # ch2 = Component(EpicsSignal, "DelayAI", write_pv="DelayAO", name="ch2", put_complete=True, kind=Kind.config)
    ch1 = Component(DummyPositioner, name="ch1")
    ch2 = Component(DummyPositioner, name="ch2")
    io = Component(DelayStatic, name="io")

    def __init__(self, *args, **kwargs):
        # Change suffix names before connecting (a bit of dynamic connections)
        self.__class__.__dict__["ch1"].suffix = kwargs["channel"][0]
        self.__class__.__dict__["ch2"].suffix = kwargs["channel"][1]
        self.__class__.__dict__["io"].suffix = kwargs["channel"]

        del kwargs["channel"]
        # Call parent to start the connections
        super().__init__(*args, **kwargs)

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        """Run a forward (pseudo -> real) calculation"""
        return self.RealPosition(ch1=pseudo_pos.delay, ch2=pseudo_pos.delay + pseudo_pos.width)

    @real_position_argument
    def inverse(self, real_pos):
        """Run an inverse (real -> pseudo) calculation"""
        return self.PseudoPosition(delay=real_pos.ch1, width=real_pos.ch2 - real_pos.ch1)


class DelayGeneratorDG645(Device):
    """DG645 delay generator

    This class implements a thin Ophyd wrapper around the Stanford Research DG645
    digital delay generator.

    Internally, the DG645 generates 8+1 signals:  A, B, C, D, E, F, G, H and T0
    Front panel outputs T0, AB, CD, EF and GH are a combination of these signals.
    Back panel outputs are directly routed signals. So signals are NOT INDEPENDENT.

    Front panel signals:
    All signals go high after their defined delays and go low after the trigger
    holdoff period, i.e. this is the trigger window. Front panel outputs provide
    a combination of these events.
    Option 1 back panel 5V signals:
    All signals go high after their defined delays and go low after the trigger
    holdoff period, i.e. this is the trigger window. The signals will stay high
    until the end of the window.
    Option 2 back panel 30V signals:
    All signals go high after their defined delays for ~100ns. This is fixed by
    electronics (30V needs quite some power). This is not implemented in the
    current device
    """

    state = Component(EpicsSignalRO, "EventStatusLI", name="status_register")
    status = Component(EpicsSignalRO, "StatusSI", name="status")

    # Front Panel
    channelT0 = Component(DelayStatic, "T0", name="T0")
    channelAB = Component(DelayPair, "", name="AB", channel="AB")
    channelCD = Component(DelayPair, "", name="CD", channel="CD")
    channelEF = Component(DelayPair, "", name="EF", channel="EF")
    channelGH = Component(DelayPair, "", name="GH", channel="GH")

    # Minimum time between triggers
    holdoff = Component(
        EpicsSignal,
        "TriggerHoldoffAI",
        write_pv="TriggerHoldoffAO",
        name="trigger_holdoff",
        kind=Kind.config,
    )
    inhibit = Component(
        EpicsSignal,
        "TriggerInhibitMI",
        write_pv="TriggerInhibitMO",
        name="trigger_inhibit",
        kind=Kind.config,
    )
    source = Component(
        EpicsSignal,
        "TriggerSourceMI",
        write_pv="TriggerSourceMO",
        name="trigger_source",
        kind=Kind.config,
    )
    level = Component(
        EpicsSignal,
        "TriggerLevelAI",
        write_pv="TriggerLevelAO",
        name="trigger_level",
        kind=Kind.config,
    )
    rate = Component(
        EpicsSignal,
        "TriggerRateAI",
        write_pv="TriggerRateAO",
        name="trigger_rate",
        kind=Kind.config,
    )

    # Command PVs
    # arm = Component(EpicsSignal, "TriggerDelayBO", name="arm", kind=Kind.omitted)

    # Burst mode
    burstMode = Component(
        EpicsSignal, "BurstModeBI", write_pv="BurstModeBO", name="burstmode", kind=Kind.config
    )
    burstConfig = Component(
        EpicsSignal, "BurstConfigBI", write_pv="BurstConfigBO", name="burstconfig", kind=Kind.config
    )
    burstCount = Component(
        EpicsSignal, "BurstCountLI", write_pv="BurstCountLO", name="burstcount", kind=Kind.config
    )
    burstDelay = Component(
        EpicsSignal, "BurstDelayAI", write_pv="BurstDelayAO", name="burstdelay", kind=Kind.config
    )
    burstPeriod = Component(
        EpicsSignal, "BurstPeriodAI", write_pv="BurstPeriodAO", name="burstperiod", kind=Kind.config
    )

    def stage(self):
        """Trigger the generator by arming to accept triggers"""
        # TODO check PV TriggerDelayBO, seems to be a bug in the IOC
        # self.arm.write(1).wait()
        super().stage()

    def unstage(self):
        """Stop the trigger generator from accepting triggers"""
        # self.arm.write(0).wait()
        super().stage()

    def burstEnable(self, count, delay, period, config="all"):
        """Enable the burst mode"""
        # Validate inputs
        count = int(count)
        assert count > 0, "Number of bursts must be positive"
        assert delay >= 0, "Burst delay must be larger than 0"
        assert period > 0, "Burst period must be positive"
        assert config in ["all", "first"], "Supported bust configs are 'all' and 'first'"

        self.burstMode.set(1).wait()
        self.burstCount.set(count).wait()
        self.burstDelay.set(delay).wait()
        self.burstPeriod.set(period).wait()

        if config == "all":
            self.burstConfig.set(0).wait()
        elif config == "first":
            self.burstConfig.set(1).wait()

    def burstDisable(self):
        """Disable the burst mode"""
        self.burstMode.set(0).wait()


# Automatically connect to test environmenr if directly invoked
if __name__ == "__main__":
    dgen = DelayGeneratorDG645("X01DA-PC-DGEN:", name="delayer")



import functools
import threading
import warnings

from ophyd.ophydobj import OphydObject
from ophyd_devices.utils.socket import SocketIO


def threadlocked(fcn):
    """Ensure that the thread acquires and releases the lock."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        lock = self._lock if hasattr(self, "_lock") else self.controller._lock
        with lock:
            return fcn(self, *args, **kwargs)

    return wrapper


class SingletonController:
    _controller_instance = None

    def __init__(self) -> None:
        self._lock = threading.RLock()

    def on(self):
        pass

    def off(self):
        pass

    def __new__(cls, *args, **kwargs):
        if not SingletonController._controller_instance:
            SingletonController._controller_instance = object.__new__(cls)
        return SingletonController._controller_instance


class Controller(OphydObject):
    _controller_instances = {}

    SUB_CONNECTION_CHANGE = "connection_change"

    def __init__(
        self,
        *,
        name=None,
        socket=None,
        attr_name="",
        parent=None,
        labels=None,
        kind=None,
    ):
        if not hasattr(self, "_initialized"):
            super().__init__(
                name=name, attr_name=attr_name, parent=parent, labels=labels, kind=kind
            )
            self._lock = threading.RLock()
            self._initialize(socket)
            self._initialized = True

    def _initialize(self, socket):
        self._connected = False
        self._set_default_values()
        self.sock = socket if socket is not None else SocketIO()

    def _set_default_values(self):
        # no. of axes controlled by each controller
        self._axis_per_controller = 8
        self._motors = [None for axis_num in range(self._axis_per_controller)]

    @property
    def connected(self):
        return self._connected

    @connected.setter
    def connected(self, value):
        self._connected = value
        self._run_subs(sub_type=self.SUB_CONNECTION_CHANGE)

    def set_motor(self, motor, axis):
        """Set the motor instance for a specified controller axis."""
        self._motors[axis] = motor

    def get_motor(self, axis):
        """Get motor instance for a specified controller axis."""
        return self._motors[axis]

    def on(self, controller_num=0):
        """Open a new socket connection to the controller"""
        if not self.connected:
            self.connected = True
        else:
            warnings.warn(f"The connection has already been established.", stacklevel=2)

    def off(self):
        """Close the socket connection to the controller"""
        self.sock.close()
        self.connected = False

    def __new__(cls, *args, **kwargs):
        socket = kwargs.get("socket")
        if not hasattr(socket, "host"):
            raise RuntimeError("Socket must specify a host.")
        if not hasattr(socket, "port"):
            raise RuntimeError("Socket must specify a port.")
        host_port = f"{socket.host}:{socket.port}"
        if host_port not in Controller._controller_instances:
            Controller._controller_instances[host_port] = object.__new__(cls)
        return Controller._controller_instances[host_port]
